#!/usr/bin/ruby

#	  Pages-Bibtex adds basic bibtex support to Apple Pages
#     Copyright (C) 2005 Thomas Counsell tamc2@cam.ac.uk and Jon Hall jgh23@open.ac.uk
# 
#     This program is free software; you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation; either version 2 of the License, or
#     (at your option) any later version.
# 
#     This program is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU General Public License for more details.
# 
#     You should have received a copy of the GNU General Public License
#     along with this program; if not, write to the Free Software
#     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

puts "
\tPages-bibtex.rb version 0.0.2, Copyright (C) 2005 Thomas Counsell, Jon Hall
\tPages-bibtex comes with ABSOLUTELY NO WARRANTY;

\tThis is free software, and you are welcome to redistribute it
\tunder certain conditions; see the source code for details.
     
\tUsage: ruby Pages-bibtex.rb documentaux_file_name.pages

"

# This is used to extend the BibItem class to change the way that references are sorted and appear.
module Style
	
	module NumberedCiteOrder
		
		def <=>( other_bib_item )
			cite_order <=> other_bib_item.cite_order
		end
		
		def citeref
			cite_order.to_s
		end
		
		def citation( type )
			case type
			when 't'; "#{shortauthor} [#{citeref}]"
			when 'author'; shortauthor
			when 'year'; year
			else ; "[#{citeref}]"
			end	
		end
	end
	
	module NumberedAlphabeticalOrder
		
		def <=>( other_bib_item )
			authors <=> other_bib_item.authors
		end
		
		def citeref
			sort_order.to_s
		end
		
		def citation( type )
			case type
			when 't'; "#{shortauthor} [#{citeref}]"
			when 'author'; shortauthor
			when 'year'; year
			else ; "[#{citeref}]"
			end	
		end
	end
	
	module Harvard
	
		attr_accessor :tail
	
		def <=>( other_bib_item )
			# This bit is just to make 1996, 1996a etc if there are duplicates
			check_for_citation_clash( other_bib_item )
			# This is the real sorting bit
			authors <=> other_bib_item.authors
		end
		
		def citeref
			"#{year}#{tail}"
		end
		
		def citation( type )
			case type
			when 't'; "#{shortauthor} (#{citeref})"
			when 'author'; shortauthor
			when 'year'; year
			else ; "(#{shortauthor} #{citeref})"
			end	
		end
		
		def check_for_citation_clash( other_bib_item )		
			return unless year == other_bib_item.year
			return unless shortauthor == other_bib_item.shortauthor
			return unless @tail == other_bib_item.tail
			if other_bib_item.tail
				other_bib_item.tail.succ!
			else
				other_bib_item.tail = 'a'
			end
		end
		
	end

	
	module ExternalBibTeX
		
		def <=>( other_bib_item )
			cite_order <=> other_bib_item.cite_order
		end
		
		def citeref
			self.bbl_citeref
		end
		
		def citation( type )
			case type
			when 't'; "#{shortauthor} [#{citeref}]"
			when 'author'; shortauthor
			when 'year'; year
			else ; "[#{citeref}]"
			end	
		end
	end
	
end

class BibFile
	
	include Enumerable
	
	def initialize( filestring = nil )
		@items = Hash.new
		parse( filestring ) if filestring
	end
	
	def each 
		@items.each { |key, bibitem| yield key, bibitem }
	end
	
	def []( key )
		return @items[ key ]
	end
	
	private
	
	def parse( filestring )
		filestring.scan(/^@[^@]*/) do |itemstring|
			bibitem = BibItem.new( itemstring )
			@items[ bibitem.key ] = bibitem
		end
	end
end

class BibItem
	
	attr_reader :type, :key, :fields, :authors, :cite_order
	attr_accessor :sort_order
	
	def initialize( itemstring )
		@fields = Hash.new
		parse itemstring
		parse_authors	
	end
	
	def []( field )
		@fields[ field.to_s.downcase ]
	end
	
	def []=( field, value )
		@fields[ field.to_s.downcase ] = value
	end
	
	def shortauthor( max = 3 )
		format_list( authors.map { |fullname| fullname.first }, max )
	end
	
	def method_missing( method, *args )
		self[ method ]
	end
	
	def cite_order=( value )
		@cite_order ||= value # So that only takes earliest citation
	end
	
	private
	
	def parse( itemstring )
		itemstring =~ /^@(.*)\{(.*?),/
		@type, @key = $1, $2 
		itemstring.scan(/(\w+)\s*=\s*\{(.*?)\}[,}]/) do |field,value|
			@fields[ field.downcase ] = value
		end
	end
	
	def parse_authors
		@authors = author ? author.split(/and/i).map { |fullname| fullname.strip.split(/,/).map { |name| name.strip } } : [["Anon",""]]
	end
	
	def format_list( elements, max = 3, tail = ' et al.' )
		return "" if elements.empty?
		return elements.first if elements.size == 1
		if elements.size > max
			return elements.first+tail
		else
			return elements[0,elements.size-1].join(', ')+' and '+elements.last
		end
	end
end

# Used if cannot find a pure ruby style file
class AuxFile
	
	def initialize( document )
		@document = document
	end
	
	def process
		create_aux_file
		run_bibtex
		load_bbl_file
		transfer_bbl_file_contents_to_citations
		clean_up_files
	end
	
	def create_aux_file
		dotauxfile = "\\bibdata\{#{@document.bibliography_file_name}\}\n"
		dotauxfile << "\\bibstyle\{#{@document.bibliography_style_name}\}\n"
		@document.bibliography.each do |citation| dotauxfile << "\\citation{#{citation.key}}\n" end
		File.open( "./#{aux_file_name}.aux", 'w' ) { |f| f.puts dotauxfile }
	end
	
	def run_bibtex
		$stderr.puts "\n\tRunning bibtex..."
		$stderr.puts `bibtex #{aux_file_name}`
		$stderr.puts "\tBibtex complete.\n\n"
	end
	
	def load_bbl_file
		@bbl = BblFile.new( IO.readlines("#{aux_file_name}.bbl").join )
	end
	
	def transfer_bbl_file_contents_to_citations
		new_citations = []
		@bbl.map do |citekey, citeref|
			bibitem = @document.bibfile[ citekey ]
			bibitem['bbl_citeref'] = citeref
			bibitem.sort_order = new_citations.size + 1
			new_citations << bibitem
		end
		@document.citations.replace( new_citations )
	end
	
	def clean_up_files
		File.delete( "#{aux_file_name}.aux" )
		File.delete( "#{aux_file_name}.bbl" )
	end

	def aux_file_name
		'pagesbibliography'
	end
end

class BblFile
	include Enumerable
	
	def initialize( filestring = nil)
		@items = []
		parse( filestring ) if filestring
	end
	
	def each 
		@items.each { |citekey, citeref| yield citekey, citeref }
	end
	
	private
	
	#I want to parse a bbl entry into citekey and citeref, if it exists
	def parse( filestring )
		filestring.scan(/\\bibitem(\[(.*?)\])?\{(.*?)\}/m) do |match|
			@items << [$3, $2 || (@items.size+1).to_s ]
		end
	end	
end

class PagesDocument
	
	attr_accessor :bibfile, :bibliography
	
	def initialize( xml )
		@xml = xml
		# Remove the insertion point (can confuse things later)
		@xml.gsub!('<sf:insertion-point/>','')
	end
	
	def citations
		@citations ||= parse_citations
	end
	
	def bibliography_file_name
		@bib_file_name ||= parse_bibliography_file_name
	end
	
	def bibliography_style_name
		@bib_style_name ||= parse_bibliography_style_name
	end
	
	def style_for( bib_type )
		styles[ bib_type.downcase ] || styles[ 'default' ] || "NO STYLES FOUND!"
	end
	
	def replace_citations
		@xml.gsub!(/\\cite(\w*)\{(.+?)\}/) do |match|
			yield $1, $2.gsub(/<.*?>/,'') # yields type, key
		end
	end
	
	def references=( bibliography_text )
		@xml.gsub!(/\\bibliography\{.*?\}/, bibliography_text )
	end
	
	def to_xml
		blank_style_templates
		@xml
	end
	
	def to_s
		to_xml
	end
	
	private
	
	def styles
		@styles ||= parse_styles
	end
	
	def parse_bibliography_file_name
		@xml =~ /\\bibliography\{(.*?)(.bib)?\}/
		$1.gsub(/<.*?>/,'') # Get rid of any xml elements
	end
	
	def parse_bibliography_style_name
		@xml =~ /\\bibliographystyle\{(.*?)\}/
		($1 && $1.gsub(/<.*?>/,'') ) || "Harvard"
	end
	
	def parse_citations
		citations = []
		@xml.scan(/\\cite\w*\{(.*?)\}/) do |match|
			citations << $1.gsub(/<.*?>/,'') # Get rid of any xml elements in the key name (it does happen!)
		end
		citations.uniq
	end
	
	def parse_styles
		styles = {}
		@xml.scan(/(<sf:p [^>]*?>)BIBLIOGRAPHY-(\w*) (.*?<\/sf:p>)/) do |match|
			styles[ $2.downcase ] = $1 + $3
			puts "\tFound style #{$2.downcase}" 
		end
		styles
	end
	
	def blank_style_templates
		@xml.gsub!(/\\bibliographystyle\{.*?\}/, '')
		@xml.gsub!(/(<sf:p [^>]*?>)BIBLIOGRAPHY-(\w*) (.*?<\/sf:p>)/,'')
	end

end

if __FILE__ == $0
		
	old_pages_file = ARGV[0]
	
	# Get rid of any trailing slashes
	old_pages_file = old_pages_file[0..-2] if old_pages_file =~ /\/$/
	
	pages_file = "#{old_pages_file[/.*(?=\.pages$)/]}-with-bibliography.pages"
	
	# Create a copy of the document
	`ditto #{old_pages_file} #{pages_file}`
	
	# Get the document out of the package	
	# Open the document
	`gunzip #{pages_file}/index.xml.gz`
	document = PagesDocument.new( IO.readlines( "#{pages_file}/index.xml" ).join )
	
	use_bibtex_to_order = false
	
	if (document.bibliography_style_name =~ /^[A-Z]+/) && Style.const_defined?(document.bibliography_style_name)
		eval("class BibItem; include Style::#{document.bibliography_style_name}; end")
	else
		$stderr.puts "\tNo ruby definition for #{document.bibliography_style_name} style"
		$stderr.puts "\tSeeing if you have BibTex and a an appropriate style file"
		class BibItem
			include Style::ExternalBibTeX
		end
		use_bibtex_to_order = true
	end
	
	# Print it out so can debug
	# require 'rexml/document'
	# xml = REXML::Document.new( document.to_xml )
	# xml.write( $stdout, 0 )
	
	# Find and open the bibliography
	$stderr.puts "\tLooking for #{document.bibliography_file_name}.bib bibliography"
	document.bibfile = BibFile.new( IO.readlines("#{document.bibliography_file_name}.bib").join )
	
	# Get the citations mentioned
	document.bibliography = document.citations.collect { |citation| document.bibfile[ citation ] }
	document.bibliography.compact! # Get rid of any missing citations
	document.bibliography.each_with_index { |citation, index| citation.cite_order = index+1 }
	
	# Sort the citations into appropriate order
	document.bibliography.sort!
	
	# Let the bibitem know their order
	document.bibliography.each_with_index { |citation, index| citation.sort_order = index+1 }
	
	# Use external bibtex if requried
	AuxFile.new( document ).process if use_bibtex_to_order

	# Create the bibliography for the end of the text
	bibliography_text = ""
	document.bibliography.each do |bibitem|
		next unless bibitem
		bibliography_text << document.style_for( bibitem.type ).gsub(/\w+/) { |m| (bibitem.send(m) || m).tr('{}','') }
	end
	
	# Replace all the citations in the text
	document.replace_citations do |type,key|
		bibitem = document.bibfile[ key ]
		unless bibitem
			$stderr.puts "\tBibliography not found for #{key}"
			key
			next
		end
		bibitem.citation( type )
	end 
	
	# Replace the \bibliography{} statement with the bibliography text
	document.references= bibliography_text
	
	# Write the document back out
	File.open( "#{pages_file}/index.xml", 'w' ) { |f| f.puts document.to_xml }
	
	# Zip it back up 
	`gzip #{pages_file}/index.xml`
	
	puts "
	Created #{pages_file}.
	WARNING: QUIT PAGES BEFORE OPENING THIS FILE
	... for some reason pages crashes if it is allready open
	"
end